/**
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 *
 *  Authors:
 *  2009-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl> 
 *
 *  2007-2009
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *
 *  2003-2007
 *     Martijn Steenbakkers <martijn@nikhef.nl>
 *     Gerben Venekamp <venekamp@nikhef.nl>
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *
 */


/*!
    \file   lcmaps_db_read.c
    \brief  the LCMAPS database reader
    \author Martijn Steenbakkers for the EU DataGrid.
*/

/*****************************************************************************
                            Include header files
******************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "lcmaps_log.h"
#include "_lcmaps_db_read.h"

/******************************************************************************
                                Definitions
******************************************************************************/
#define MAXDBENTRIES 250 /*!< maximum number of LCMAPS database entries \internal */
#define MAXPAIRS     2   /*!< maximum number of variable-value pairs that will be parsed per line
                            \internal */

#define WHITESPACE_CHARS " \t\n" /*!< \internal */
#define QUOTING_CHARS    "\"" /*!< \internal */
#define ESCAPING_CHARS   "\\" /*!< \internal */
#define COMMENT_CHARS    "#" /*!< \internal */

/*! Characters separating variable-value pairs in the lcmaps database file \internal */
#define PAIR_SEP_CHARS ","
/*! Characters separating variables from values \internal */
#define VARVAL_SEP_CHARS "="

/*
 * Characters that terminate pairs, values and variables in the lcmaps database file. This
 * is a combination of whitespace and separators.
 */
/*!
    Characters that terminate pairs in the lcmaps database file. This
    is a combination of whitespace and separators. \internal
 */
#if 0
#define PAIR_TERMINATOR_CHARS PAIR_SEP_CHARS WHITESPACE_CHARS
#endif
/*!
    Characters that terminate variables and values in the lcmaps database file. This
    is a combination of whitespace and separators. \internal
 */
#if 0
#define VARVAL_TERMINATOR_CHARS VARVAL_SEP_CHARS WHITESPACE_CHARS
#endif

#ifndef NUL
#define NUL '\0' /*!< \internal */
#endif

/******************************************************************************
                          Module specific prototypes
******************************************************************************/
static int lcmaps_db_read_entries(FILE *);
static int lcmaps_db_parse_line(char *, lcmaps_db_entry_t **);
static int lcmaps_db_parse_pair(char *, char **, char **);
static int lcmaps_db_parse_string(char **);

/******************************************************************************
                       Define module specific variables
******************************************************************************/
static lcmaps_db_entry_t *  lcmaps_db_list=NULL; /*!< list of database entries */

/******************************************************************************
    Function: lcmaps_db_read
    documentation in _lcmaps_db_read.h
******************************************************************************/
/*!
    \fn lcmaps_db_read(
        char * lcmaps_db_fname
        )
    \brief Read database from file.
    \param lcmaps_db_fname database file.
    \return a pointer to the database list
    \internal
*/
lcmaps_db_entry_t ** lcmaps_db_read(
        char * lcmaps_db_fname
)
{
    FILE * lcmaps_db_fhandle;
    int    no_entries;

    lcmaps_db_fhandle = fopen(lcmaps_db_fname, "r");
    if (lcmaps_db_fhandle == NULL)
    {
        /* Cannot open file */
        return NULL;
    }

    no_entries=lcmaps_db_read_entries(lcmaps_db_fhandle);
    if (no_entries < 0)
    {
        lcmaps_log(LOG_ERR,"lcmaps.mod-lcmaps_db_read(): Parse error in line %d of %s\n",
            -(no_entries), lcmaps_db_fname);
        fclose(lcmaps_db_fhandle);
        return NULL;
    }
    else if (no_entries > MAXDBENTRIES)
    {
        lcmaps_log(LOG_ERR,"lcmaps.mod-lcmaps_db_read(): Too many entries found in %s\n",
            lcmaps_db_fname);
        lcmaps_log(LOG_ERR,"lcmaps.mod-lcmaps_db_read(): Only the first %d entries are used\n",
            MAXDBENTRIES);
    }
    fclose(lcmaps_db_fhandle);

    return &lcmaps_db_list;
}

/*!
    \fn lcmaps_db_read_entries(
        FILE * dbstream
        )
    \brief Read db entries from stream and fill a lsit of db entries
    \param dbstream database stream
    \return the number of entries found (failure --> negative number)
    \internal
*/
static int lcmaps_db_read_entries(
        FILE * dbstream
)
{
    char               line[1024];
    int                nlines=0;
    int                no_entries=0;
    lcmaps_db_entry_t *  entry=NULL;

    nlines=0;
    no_entries=0;
/*    lcmaps_db_fill_entry(no_entries,NULL); */
    while (fgets(line, (int)sizeof(line), dbstream))
    {
        ++nlines;

        if (! lcmaps_db_parse_line(line, &entry))
        {
            /* parse error, return line number */
            if (entry != NULL) free(entry);
            return -(nlines);
        }
        if (entry != NULL)
        {
            lcmaps_log_debug(3,"line %d: %s, %s\n",nlines,entry->pluginname,entry->pluginargs);
            /* entry found */
            ++no_entries;
            if (no_entries > MAXDBENTRIES)
            {
                if (entry != NULL) free(entry);
                return no_entries;
            }
            if ( lcmaps_db_fill_entry(&lcmaps_db_list, entry)==NULL )
            {
                /* error filling lcmaps_db */
                if (entry != NULL) free(entry);
                return -(nlines);
            }
            if (entry != NULL) free(entry);
            entry=NULL;
        }
        else
        {
            /* no entry found, but no error */
            continue;
        }
    }
    if (entry != NULL) free(entry);
    return no_entries;
}

/******************************************************************************
    Function: lcmaps_db_fill_entry
    documentation in _lcmaps_db_read.h
******************************************************************************/
/*!
    \fn lcmaps_db_fill_entry(
        lcmaps_db_entry_t ** list,
        lcmaps_db_entry_t *  entry
        )
    \brief Add a database entry to a list.
    \param list database list (array of database entry pointers)
    \param entry the database entry to be added
    \return a pointer to the newly created database entry in the list or NULL (error)
    \internal
*/
lcmaps_db_entry_t * lcmaps_db_fill_entry(
        lcmaps_db_entry_t ** list,
        lcmaps_db_entry_t * entry
)
{
    lcmaps_db_entry_t * plist;

    if (entry == NULL)
    {
        lcmaps_log(LOG_ERR,"lcmaps.mod-lcmaps_db_fill_entry(): error null entry\n");
        return NULL;
    }

    if (!(*list))
    {
        lcmaps_log_debug(2,"lcmaps.mod-lcmaps_db_fill_entry(): creating first list entry\n");
        *list=plist=(lcmaps_db_entry_t *)malloc(sizeof(lcmaps_db_entry_t));
    }
    else
    {
        lcmaps_log_debug(2,"lcmaps.mod-lcmaps_db_fill_entry(): creating new list entry\n");
        plist=*list;
        while (plist->next) plist=plist->next;
        plist=plist->next=(lcmaps_db_entry_t *)malloc(sizeof(lcmaps_db_entry_t));
    }
    if (!plist)
    {
        lcmaps_log(LOG_ERR,"lcmaps.mod-lcmaps_db_fill_entry(): error creating new list entry\n");
        return NULL;
    }
    plist->next=NULL;

    if (entry->pluginname != NULL)
    {
        strncpy(plist->pluginname,entry->pluginname,LCMAPS_MAXPATHLEN);
        (plist->pluginname)[LCMAPS_MAXPATHLEN]=NUL;
    }
    else
        strncpy(plist->pluginname,"",LCMAPS_MAXPATHLEN);

    if (entry->pluginargs != NULL)
    {
        strncpy(plist->pluginargs,entry->pluginargs,LCMAPS_MAXARGSTRING);
        (plist->pluginargs)[LCMAPS_MAXARGSTRING]=NUL;
    }
    else
        strncpy(plist->pluginargs,"",LCMAPS_MAXARGSTRING);

    return plist;
}

/*!
    \fn lcmaps_db_parse_line(
        char * line,
        lcmaps_db_entry_t ** entry
        )
    \brief Parses database line and fills database structure
    \param line database line
    \param entry pointer to a pointer to a database structure (can/should be freed afterwards)
    \retval 1 succes.
    \retval 0 failure.
    \internal
*/
static int lcmaps_db_parse_line(
        char * line,
        lcmaps_db_entry_t ** entry
)
{
    char *            var_val_pairs[MAXPAIRS];
    char *            var;
    char *            val;
    int               ipair;
    int               no_pairs;
    size_t            len;
    lcmaps_db_entry_t * tmp_entry=NULL;

    /* Check arguments */
    if ((line == NULL) || (*entry != NULL))
    {
        lcmaps_log(LOG_ERR,"lcmaps.mod-lcmaps_db_parse_line(): something wrong with arguments\n");
        goto error;
    }

    /* Skip over leading whitespace */
    line += strspn(line, WHITESPACE_CHARS);

    /* Check for comment at start of line and ignore line if present */
    if (strchr(COMMENT_CHARS, *line) != NULL) 
    {
        /* Ignore line, return NULL entry. */
        *entry=NULL;
        return 1;
    }

    /* Check for empty line */
    if (*line == NUL)
    {
        /* Empty line, return NULL entry. */
        *entry=NULL;
        return 1;
    }

    /* Look for variable-value pairs by checking the PAIR_SEP_CHARS */
    ipair=0;
    len=0;
    while (*line != NUL)
    {
        len=strcspn(line, PAIR_SEP_CHARS);
        if (len > 0)
        {
            var_val_pairs[ipair] = line;
            ipair++;
            line+=len;
            if (*line == NUL)
            {
                /* probably end of line */
                continue;
            }
            if (strchr(PAIR_SEP_CHARS, *line) != NULL)
            {
                *line=NUL;
                line += 1;
            }
        }
        else
        {
            /* len = 0, so *line=PAIR_SEP_CHARS */
            line += 1;
        }
        line += strspn(line, WHITESPACE_CHARS);
    }
    no_pairs=ipair;
    if (no_pairs)
    {
        tmp_entry=malloc(sizeof(*tmp_entry));
        if (tmp_entry == NULL)
        {
            lcmaps_log(LOG_ERR,"lcmaps.mod-lcmaps_db_parse_line(): error allocating memory\n");
            goto error;
        }

        /* Initialize tmp_entry */
        (tmp_entry->pluginname)[0] = NUL;
        (tmp_entry->pluginargs)[0] = NUL;
        for (ipair=0; ipair < no_pairs; ipair++)
        {
            int rc=0;
            lcmaps_log_debug(3,"pair %d:%s-endpair\n",ipair, var_val_pairs[ipair]);
            rc=lcmaps_db_parse_pair(var_val_pairs[ipair], &var, &val);
            if (! rc)
            {
                /* error parsing variable-value pair */
                lcmaps_log(LOG_ERR,"lcmaps.mod-lcmaps_db_parse_line(): error parsing variable-value pair\n");
                goto error;
            }
            lcmaps_log_debug(3,"var: %s, value: %s\n",var,val);

            if (strncmp(var,"pluginname",strlen("pluginname")) == 0)
            {
                /* found plugin name */
                strncpy(tmp_entry->pluginname,val,LCMAPS_MAXPATHLEN);
                (tmp_entry->pluginname)[LCMAPS_MAXPATHLEN]=NUL;
            }
            else if (strncmp(var,"pluginargs",strlen("pluginargs")) == 0)
            {
                /* found plugin database */
                strncpy(tmp_entry->pluginargs,val,LCMAPS_MAXARGSTRING);
                (tmp_entry->pluginargs)[LCMAPS_MAXARGSTRING]=NUL;
            }
            else
            {
                /* unknown option: do nothing */
            }
        }
    }

    /* succes */
    *entry=tmp_entry;
    return 1;

  error:
    if (tmp_entry != NULL) free(tmp_entry);
    return 0;
}

/*!
    \fn lcmaps_db_parse_pair(
        char * pair,
        char ** pvar,
        char ** pval
        )
    \brief Parses a database variable-value pair and returns the variable name and its value
    \param pair string containing the pair
    \param pvar pointer to the variable string
    \param pval pointer to the value string
    \retval 1 succes.
    \retval 0 failure.
    \internal
*/
static int lcmaps_db_parse_pair(
        char * pair,
        char ** pvar,
        char ** pval
)
{
    size_t len;
    char * var;
    char * val;

    if (pair == NULL)
    {
        lcmaps_log(LOG_ERR,"lcmaps.mod-lcmaps_db_parse_pair(): cannot parse empty pair\n");
        return 0;
    }

    /* Skip over leading whitespace */
    pair += strspn(pair, WHITESPACE_CHARS);
    if (*pair == NUL)
    {
        lcmaps_log(LOG_ERR,"lcmaps.mod-lcmaps_db_parse_pair(): cannot parse empty pair\n");
        return 0;
    }

    var=pair;
    len=strcspn(pair, VARVAL_SEP_CHARS);
    pair+=len;
    if (*pair == NUL)
    {
        /* Cannot find '=' */
        lcmaps_log(LOG_ERR,"lcmaps.mod-lcmaps_db_parse_pair(): cannot find =\n");
        return 0;
    }

    if (strchr(VARVAL_SEP_CHARS, *pair) != NULL)
    {
        /* Found ' var =' and replace '=' with NUL*/
        *pair=NUL;
        if (! lcmaps_db_parse_string(&var))
        {
            /* error parsing variable name */
            return 0;
        }
        pair+=1;
        if (*pair==NUL)
        {
            /* No value found */
/*            val=NULL; */
            val=pair;
            *pvar=var;
            *pval=val;
            return 1;
        }
        else
        {
            /* Skip over leading whitespace */
            pair += strspn(pair, WHITESPACE_CHARS);
            if (*pair == NUL)
            {
                /* No value found */
/*                val=NULL; */
                val=pair;
                *pvar=var;
                *pval=val;
                return 1;
            }
            val=pair;
            if (! lcmaps_db_parse_string(&val))
            {
                /* error parsing value */
                return 0;
            }
        }
    }
    else
    {
        /* Cannot find '=' */
        lcmaps_log(LOG_ERR,"lcmaps.mod-lcmaps_db_parse_pair(): cannot find =\n");
        return 0;
    }

    /* success */
    *pvar=var;
    *pval=val;
    return 1;
}

/*!
    \fn lcmaps_db_parse_string(
        char ** pstring
        )
    \brief Takes a string and removes prepending and trailing spaces and quotes (unless escaped)
    \param pstring pointer to a pointer to a char
    \retval 1 succes.
    \retval 0 failure.
    \internal
*/
static int lcmaps_db_parse_string(
        char ** pstring
)
{
    char * end;
    char * string;

    string=*pstring;

    if (*string==NUL)
    {
        lcmaps_log(LOG_ERR,"lcmaps.mod-lcmaps_db_parse_string(): error parsing null string\n");
        return 0;
    }

    /* Is string quoted ? */
    if (strchr(QUOTING_CHARS, *string) != NULL)
    {
        /*
         * Yes, skip over opening quote and look for unescaped
         * closing double quote
         */
        string++;
        end=string;
        do
        {
            end += strcspn(end, QUOTING_CHARS);
            if (*end == NUL)
            {
                lcmaps_log(LOG_ERR,"lcmaps.mod-lcmaps_db_parse_string(): missing closing quote\n");
                return 0; /* Missing closing quote */
            }

            /* Make sure it's not escaped */
        }
        while (strchr(ESCAPING_CHARS, *(end - 1)) != NULL);
    }
    else
    {
        end = string + strcspn(string, WHITESPACE_CHARS);
    }
    *end=NUL;
    *pstring=string;

    return 1;
}

/******************************************************************************
    Function: lcmaps_db_clean_list
    documentation in _lcmaps_db_read.h
******************************************************************************/
/*!
    \fn lcmaps_db_clean_list(
        lcmaps_db_entry_t ** list
        )
    \brief Clean/remove the database list
    \param list pointer to the database list
    \retval 0 succes.
    \retval 1 failure.
    \internal
*/
int lcmaps_db_clean_list(
        lcmaps_db_entry_t ** list
)
{
    lcmaps_db_entry_t * entry;

    entry=*list;
    while (entry)
    {
        lcmaps_db_entry_t * next_entry;
        lcmaps_log_debug(2,"cleaning db entry for module %s\n",entry->pluginname);
        next_entry=entry->next;
        free(entry);
        entry=next_entry;
    }
    *list=entry=NULL;
    return 0;
}

/******************************************************************************
    Function: lcmaps_db_clean
    documentation in _lcmaps_db_read.h
******************************************************************************/
/*!
    \fn lcmaps_db_clean()
    \brief Clean/remove the database structure.
    \retval 0 succes
    \retval 1 failure
    \internal
*/
int lcmaps_db_clean(void)
{
    if(lcmaps_db_clean_list(&lcmaps_db_list) != 0)
    {
        lcmaps_log(LOG_ERR,"lcmaps.mod-lcmaps_db_clean() error: could not clean list\n");
        return 1;
    }
    return 0;
}

/******************************************************************************
CVS Information:
    $Source: /srv/home/dennisvd/svn/mw-security/lcmaps/src/pluginmanager/lcmaps_db_read.c,v $
    $Date: 2014-02-26 16:01:45 +0100 (Wed, 26 Feb 2014) $
    $Revision: 17529 $
    $Author: msalle $
******************************************************************************/
